//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "Unit1.h"

// jhaley: Include TWAIN standard header (try with v2.1 header first; if doesn't work, try v1.9)
#include "twain.h"
#include <string>
#include <vector>
#include <map>
#include "string.h"
#include "stdio.h"
#include "stdlib.h"
using std::vector;
using std::string;
using std::map;
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;

// jhaley: messy globals for TWAIN stuff. In Prometheus, we should probably have a clean OO approach instead.
// The TWAIN architecture is naturally object-oriented. It has:
// * A data source manager (DSM)
// * Data sources (usually only one at a time) (DS)
// * The application
// * A global state (ranging from 1 to 7)
// * Device Caps
// * ETC. (only very advanced uses will need anything more than the above)
// Each of these could be encapsulated, but this app is just for testing purposes, to see how this all gets along with
// Borland and VCL. (So far so good)

DSMENTRYPROC pDSM_Entry; // The DSM_Entry routine loaded from the DLL
HANDLE       hDSMLib;    // Handle to the DSM library (DLL)

// State vars
bool twain_loaded          = false; // DSM loaded successfully     - In state 2 or higher
bool twain_opened          = false; // DSM opened successfully     - In state 3 or higher
bool twain_source_opened   = false; // DS opened successfully      - In state 4 or higher
bool twain_caps_negotiated = false; // Capabilities negotiated       * during state 4
bool twain_source_enabled  = false; // DS currently enabled        - In state 5 or higher
bool twain_transfer_ready  = false; // Transfer is ready           - In state 6 or higher

// Identity Structures
TW_IDENTITY AppId; // The Application ID structure, sent to the TWAIN DSM - must be run-time-constant once initialized
TW_IDENTITY SrcId; // The Source ID structure of the currently selected source (not necessarily open)
TW_IDENTITY SelId; // The Source ID structure of the currently opened source (not necessarily the selected source!)

// Other data
TW_INT16     twain_negotiated_image_count = 0; // Negotiated image count with selected source, in case needed

map<int, string>     imageFileNames; // Filenames
int                  currentImage;   // Index of image being displayed

//---------------------------------------------------------------------------
// jhaley: Bitmap management routines start here vvvvv
//---------------------------------------------------------------------------
// WriteBitmapFile
// Write the HBITMAP out to a binary .bmp file and throw the generated filename onto the imageFileNames vector.
// I tried just assigning the handles to the TImage::Picture's Bitmap but that does not work, for whatever reason.
// It might be possible to rewrite this to use a memory stream with the LoadFromStream method and avoid writing out
// files. Then again it might be useful to have the file around.
static void WriteBitmapFile(HBITMAP hBmp, int num)
{
  char filename[MAX_PATH];
  ZeroMemory(filename, MAX_PATH);

  PBITMAPINFOHEADER dib = (PBITMAPINFOHEADER)GlobalLock(hBmp);
  if (!dib)
    return;

  sprintf(filename, "scan_%06d.bmp", num);

  FILE *f = fopen(filename, "wb");
  if (f)
  {
    DWORD palette_size = 0;

    switch (dib->biBitCount)
    {
    case 1:  palette_size = 2;   break;
    case 8:  palette_size = 256; break;
    case 24: break;
    default: goto lbl_error; // ick.
    }

    if (dib->biSizeImage == 0)
    {
      dib->biSizeImage = ((((dib->biWidth * dib->biBitCount) + 31) & ~31) / 8) * dib->biHeight;
      if (dib->biCompression != 0)
        dib->biSizeImage = (dib->biSizeImage * 3) / 2;
    }

    int image_size = dib->biSizeImage + (sizeof(RGBQUAD) * palette_size) + sizeof(BITMAPINFOHEADER);
    BITMAPFILEHEADER fh = { 0 };
    fh.bfType = ((WORD)('M' << 8) | 'B');
    fh.bfSize = image_size + sizeof(BITMAPFILEHEADER);
    fh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + (sizeof(RGBQUAD) * palette_size);

    fwrite(&fh, 1, sizeof(BITMAPFILEHEADER), f);
    fwrite(dib, 1, image_size, f);
    fclose(f);

    // Write was fully successful so save this filename
    imageFileNames[num] = string(filename);
  }

lbl_error:
  if (f)
    fclose(f);

  GlobalUnlock(hBmp);
  GlobalFree(hBmp);
}
//---------------------------------------------------------------------------
static void DisplayBitmap(TImage &image, int i)
{
  currentImage = i;
  if (imageFileNames[i] != "")
    image.Picture->Bitmap->LoadFromFile(imageFileNames[i].c_str());
}
//---------------------------------------------------------------------------
static void DisplayNextBitmap(TImage &image)
{
  if (imageFileNames[currentImage+1] != "")
  {
    DisplayBitmap(image, currentImage+1);
  }
}
//---------------------------------------------------------------------------
static void DisplayPrevBitmap(TImage &image)
{
  if (imageFileNames[currentImage-1] != "")
  {
    DisplayBitmap(image, currentImage-1);
  }
}
//---------------------------------------------------------------------------
// End of bitmap management routines ^^^^^
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// jhaley: TWAIN prototype interface code starts here vvvvv
//---------------------------------------------------------------------------
// LoadSourceManager
// Brings the application into State 2 by loading the TWAIN source manager
static void LoadSourceManager()
{
  char winPath[MAX_PATH];

  ZeroMemory(winPath, sizeof(winPath));

  // Check for existence of the TWAIN_32.DLL file in the Windows directory
  GetWindowsDirectory(winPath, sizeof(winPath));
  if (winPath[(lstrlen(winPath) - 1)] != '\\')
    lstrcat(winPath, "\\");
  lstrcat(winPath, "TWAIN_32.DLL"); // FIXME: insecure strcat calls, blech (use std::string)

  if ((hDSMLib = LoadLibrary(winPath)) != NULL)
  {
    if ((pDSM_Entry = (DSMENTRYPROC)GetProcAddress(hDSMLib, MAKEINTRESOURCE(1))) == NULL)
    {
      // If we couldn't get the function, unload the library.
      FreeLibrary(hDSMLib);
      Application->MessageBox("Failed to load TWAIN_32.DLL!", "Error", MB_OK);
    }
    else
      twain_loaded = true;
  }
}
//---------------------------------------------------------------------------
// OpenSourceManager
// Brings the application into State 3 by opening the TWAIN source manager
static void OpenSourceManager(TForm *form)
{
  ZeroMemory(&AppId, sizeof(AppId));

  // Initialize the AppId object
  AppId.Id = 0; // Will be set to a globally unique id by the DSM
  AppId.Version.MajorNum = 1;
  AppId.Version.MinorNum = 0;
  AppId.Version.Language = TWLG_ENGLISH_USA;
  AppId.Version.Country  = TWCY_USA;
  lstrcpy(AppId.Version.Info, "v1.0"); // Is this what's really expected here? Does it really matter?
  AppId.ProtocolMajor = TWON_PROTOCOLMAJOR;
  AppId.ProtocolMinor = TWON_PROTOCOLMINOR;
  AppId.SupportedGroups = DG_IMAGE | DG_CONTROL;
  lstrcpy(AppId.Manufacturer,  "Absentee Shawnee Tribe");
  lstrcpy(AppId.ProductFamily, "Health IT Test Apps");
  lstrcpy(AppId.ProductName,   "TWAIN Test Application");

  TW_UINT16 rc;
  HWND handle = form->Handle; // Spec is a bit vague but this seems to be correct.

  //                 pOrigin, pDest, DG,         DAT,        MSG,         pData
  rc = (*pDSM_Entry)(&AppId,  NULL,  DG_CONTROL, DAT_PARENT, MSG_OPENDSM, (TW_MEMREF)&handle);

  if (rc != TWRC_SUCCESS)
  {
    // Rut-row, Raggy!
    Application->MessageBox("Failed to open TWAIN source manager!", "Error", MB_OK);
  }
  else
    twain_opened = true;

  // Init source identity structure to "default source" state
  ZeroMemory(&SrcId, sizeof(SrcId));
}
//---------------------------------------------------------------------------
// SelectSource
// Called in the test app by a click to the "Select source..." menu item. Displays the DSM's source selection UI and
// records the selection made.
static void SelectSource()
{
  if (twain_opened) // Must be in state 3, at least.
  {
    TW_UINT16 rc;
    TW_IDENTITY TempSrcId;
    string msg;

    ZeroMemory(&TempSrcId, sizeof(TempSrcId));

    rc = (*pDSM_Entry)(&AppId, NULL, DG_CONTROL, DAT_IDENTITY, MSG_USERSELECT, (TW_MEMREF)&TempSrcId);

    switch (rc)
    {
    case TWRC_SUCCESS:
      // Copy TempSrcId to SrcId
      SrcId = TempSrcId;
      msg = string("You selected the '") + SrcId.ProductName + "' device.";
      Application->MessageBox(msg.c_str(), "Success!", MB_OK);
      break;
    case TWRC_CANCEL:
      // Do nothing.
      break;
    case TWRC_FAILURE:
      Application->MessageBox("Failed to select a source!", "Error", MB_OK);
      break;
    }
  }
  else
    Application->MessageBox("Cannot select a source because the TWAIN DSM is not open.", "Error", MB_OK);
}
//---------------------------------------------------------------------------
// OpenSource
// Part 1 of the operation performed for image acquisition. Transitions the application into state 4 by opening a
// source. Only one source should be open at a time, and for the sake of sanity I am keeping a source open only during
// the process of acquisition.
// Note that if a source has not been manually selected beforehand, the "default" source will be opened. This is the
// usual behavior encouraged (though not necessarily required) of all TWAIN-compliant applications.
static void OpenSource()
{
  TW_UINT16 rc;

  if (!twain_source_opened) // Must not already be in state 4. Only one source can be open at a time.
  {
    // Try to open the selected (or default) source
    rc = (*pDSM_Entry)(&AppId, NULL, DG_CONTROL, DAT_IDENTITY, MSG_OPENDS, (TW_MEMREF)&SrcId);

    if (rc == TWRC_SUCCESS)
    {
      // The source is now open. Copy SrcId to SelId, as we must continue using this source identity information as long
      // as we remain in state 4 or higher, even if the user selects a different device during that period.
      SelId = SrcId;
      twain_source_opened = true;

      // A source, being just opened, has not had capabilities negotiated yet.
      twain_caps_negotiated = false;
      twain_negotiated_image_count = 0;
    }
    else
      Application->MessageBox("Could not open the selected source!", "Error", MB_OK);
  }
}
//---------------------------------------------------------------------------
// CheckCapability
// If the device cannot set the CAP_XFERCOUNT capability exactly it will attempt to negotiate an acceptable value. This
// is a subroutine for NegotiateCapabilities. This could probably be generalized for capability negotiation in general.
static bool CheckCapability(TW_INT16 &count)
{
  TW_CAPABILITY twCapability;
  TW_UINT16     rc;
  pTW_ONEVALUE  pval;

  twCapability.Cap = CAP_XFERCOUNT;
  twCapability.ConType = TWON_DONTCARE16; // Source will specify
  twCapability.hContainer = NULL;         // Source allocates and fills container, but we still have to free it.

  rc = (*pDSM_Entry)(&AppId, &SelId, DG_CONTROL, DAT_CAPABILITY, MSG_GET, (TW_MEMREF)&twCapability);
  if(twCapability.hContainer)
  {
    bool ret;
    pval = (pTW_ONEVALUE)GlobalLock(twCapability.hContainer);
    if(pval)
    {
      count = pval->Item;
      ret = true;
    }
    else
    {
      Application->MessageBox("Could not lock source value!", "Error", MB_OK);
      ret = false;
    }
    GlobalFree((HANDLE)twCapability.hContainer);
    return ret;
  }
  else
  {
    // I think only a really buggy source would do this, but I'm being as safe as possible. Screwing with GlobalAlloc
    // is asking for a hard crash for sure.
    Application->MessageBox("Source returned NULL cap container!", "Error", MB_OK);
    return false;
  }
}
//---------------------------------------------------------------------------
// NegotiateCapabilities
// Part 2 of the operation performed for image acquisition. With a source opened, we must negotiate with the source for
// the supported/desired capabilities.
static void NegotiateCapabilities()
{
  if (twain_source_opened) // Must be in state 4
  {
    TW_CAPABILITY twCapability;
    TW_INT16      count;
    TW_STATUS     twStatus;
    TW_UINT16     rc;
    pTW_ONEVALUE  pval;

    // Setup for MSG_GET for CAP_XFERCOUNT
    twCapability.Cap = CAP_XFERCOUNT;
    twCapability.ConType = TWON_ONEVALUE;
    twCapability.hContainer = GlobalAlloc(GHND, sizeof(TW_ONEVALUE));

    if (twCapability.hContainer)
    {
      pval = (pTW_ONEVALUE)GlobalLock(twCapability.hContainer);
      if (pval)
      {
        pval->ItemType = TWTY_INT16;
        pval->Item     = 1; // This app will accept only 1 image
        GlobalUnlock(twCapability.hContainer);

        // Set the CAP_XFERCOUNT
        rc = (*pDSM_Entry)(&AppId, &SelId, DG_CONTROL, DAT_CAPABILITY, MSG_SET, (TW_MEMREF)&twCapability);

        // Check return codes
        if (rc == TWRC_SUCCESS)
        {
          twain_caps_negotiated = true;
          twain_negotiated_image_count = 1; // TODO: It may not even really be necessary to track this... I'll know in a bit.
        }
        else if (rc == TWRC_CHECKSTATUS) // Value could not be set exactly
        {
          if (CheckCapability(count)) // Check the actual capability offered; we'll try to work around it, I guess.
          {
            twain_caps_negotiated = true;
            twain_negotiated_image_count = count; // TODO: It may not even really be necessary to track this... I'll know in a bit.
          }
        }
        else if (rc == TWRC_FAILURE) // Woops.
        {
          // Check condition code (TODO: this should probably end up being its own function or method, for error
          // reporting purposes).
          rc = (*pDSM_Entry)(&AppId, &SelId, DG_CONTROL, DAT_STATUS, MSG_GET, (TW_MEMREF)&twStatus);

          switch (twStatus.ConditionCode)
          {
          case TWCC_BADCAP:          // NB: for sources compliant to less than v1.7 only.
          case TWCC_CAPUNSUPPORTED:
          case TWCC_CAPBADOPERATION:
          case TWCC_CAPSEQERROR:
            // Not supported, but should be according to the spec...
            Application->MessageBox("Source is not TWAIN-compliant! Can't set CAP_XFERCOUNT.", "Error", MB_OK);
            break;
          case TWCC_BADDEST:
            // Source not open?!? Shouldn't happen.
            Application->MessageBox("The source is not open!", "Error", MB_OK);
            break;
          case TWCC_BADVALUE:
            // Value is out of range?!?
            Application->MessageBox("Source says requested CAP_XFERCOUNT is invalid!", "Error", MB_OK);
            break;
          case TWCC_SEQERROR:
            // Sequence error?!? Also shouldn't happen.
            Application->MessageBox("The application is not in a valid state!", "Error", MB_OK);
            break;
          }
        }
      }
      else
        Application->MessageBox("Devcaps negotiation failed in GlobalLock!", "Error", MB_OK);

      // Release the value container
      GlobalFree((HANDLE)twCapability.hContainer);
    }
    else
      Application->MessageBox("Devcaps negotiation failed in GlobalAlloc!", "Error", MB_OK);
  }
}
//---------------------------------------------------------------------------
// EnableSource
// Part 3 of the image acquisition process. EnableSource puts the application into state 5.
static void EnableSource(TForm *form)
{
  if (twain_source_opened) // Must be in state 4
  {
    TW_USERINTERFACE twInterface;
    TW_UINT16 rc;

    twInterface.ShowUI  = TRUE;
    twInterface.ModalUI = FALSE;
    twInterface.hParent = form->Handle;

    rc = (*pDSM_Entry)(&AppId, &SelId, DG_CONTROL, DAT_USERINTERFACE, MSG_ENABLEDS, (TW_MEMREF)&twInterface);
    if (rc == TWRC_SUCCESS)
      twain_source_enabled = true; // Now in state 5; message forwarding will begin immediately (see AppMessage)
    else
      Application->MessageBox("Failed to enable source!", "Error", MB_OK);
  }
}
//---------------------------------------------------------------------------
// UpdateImageInfo
// Subroutine for SetupAndTransferImage - gets an imageInfo structure
static bool UpdateImageInfo(TW_IMAGEINFO *imageInfo)
{
  TW_UINT16 rc;

  ZeroMemory(imageInfo, sizeof(*imageInfo));

  rc = (*pDSM_Entry)(&AppId, &SelId, DG_IMAGE, DAT_IMAGEINFO, MSG_GET, (TW_MEMREF)imageInfo);

  return (rc == TWRC_SUCCESS);
}
//---------------------------------------------------------------------------
// DoAbortXfer
// Subroutine for SetupAndTransferImage, for error handling.
static TW_UINT16 DoAbortXfer()
{
  TW_UINT16 rc = TWRC_SUCCESS;
  TW_PENDINGXFERS pendingXfers;

  ZeroMemory(&pendingXfers, sizeof(pendingXfers));

  // Try to end the current transfer
  rc = (*pDSM_Entry)(&AppId, &SelId, DG_CONTROL, DAT_PENDINGXFERS, MSG_ENDXFER, (TW_MEMREF)&pendingXfers);
  if(pendingXfers.Count != 0)
  {
    // Abort remaining transfers
    ZeroMemory(&pendingXfers, sizeof(pendingXfers));
    rc = (*pDSM_Entry)(&AppId, &SelId, DG_CONTROL, DAT_PENDINGXFERS, MSG_RESET, (TW_MEMREF)&pendingXfers);
  }

  return rc;
}
//---------------------------------------------------------------------------
// SetupAndTransferImage
// Part 4 of the acquisition process. This is called when the source sends an event back to the application via the
// DSM message loop handler (AppMessage) to indicate that it is prepared to send data. It will wait for acknowledgement.
// This routine places the application into state 7, but in a transitory manner.
static void SetupAndTransferImage(TForm1 *form)
{
  TW_UINT16 rc;
  bool have_pending_xfers = true;
  static int xfer_count = 0;

  if (twain_transfer_ready)
    return; // ????

  // We are now in state 6.
  twain_transfer_ready = true;

  while (have_pending_xfers)
  {
    xfer_count++;

    TW_IMAGEINFO currInfo;
    HBITMAP      currHandle = 0;

    if(!UpdateImageInfo(&currInfo))
      break; // error

    // get image handle - begins state 7!
    rc = (*pDSM_Entry)(&AppId, &SelId, DG_IMAGE, DAT_IMAGENATIVEXFER, MSG_GET, (TW_MEMREF)&currHandle);

    if (rc == TWRC_XFERDONE)
    {
      TW_PENDINGXFERS pendingXfers;
      ZeroMemory(&pendingXfers, sizeof(pendingXfers));

      // Write out the image
      WriteBitmapFile(currHandle, xfer_count);

      rc = (*pDSM_Entry)(&AppId, &SelId, DG_CONTROL, DAT_PENDINGXFERS, MSG_ENDXFER, (TW_MEMREF)&pendingXfers);
      if (rc == TWRC_SUCCESS)
      {
        if (pendingXfers.Count == 0)
          have_pending_xfers = false; // Done, end loop.
      }
      else // error! (We'd probably want to log this in Prometheus; a message box is not a good idea while the source is active)
        have_pending_xfers = false;
    }
    else if (rc == TWRC_CANCEL)
      break; // error, get out!
    else if (rc == TWRC_FAILURE)
      break; // error, get out!
  }

  if (have_pending_xfers) // left the loop due to an error?
    rc = DoAbortXfer();

  DisplayBitmap(*(form->imgBitmap), xfer_count);

  // return to state 5.
  twain_transfer_ready = false;
}
//---------------------------------------------------------------------------
// DisableSource
// Shuts down the source. Transitions back to state 4.
static void DisableSource()
{
  if (twain_source_enabled) // Must be in state 5
  {
    TW_UINT16 rc;
    TW_USERINTERFACE foo; // not used for anything

    ZeroMemory(&foo, sizeof(foo));

    rc = (*pDSM_Entry)(&AppId, &SelId, DG_CONTROL, DAT_USERINTERFACE, MSG_DISABLEDS, (TW_MEMREF)&foo);

    if (rc == TWRC_SUCCESS)
    {
      // Now back in state 4
      twain_source_enabled = false;
    }
    // Otherwise an error, but this is semi-unrecoverable except at shutdown. Would want to log in Prometheus.
  }
}
//---------------------------------------------------------------------------
// CloseSource
// Closes the source. Transitions back to state 3.
static void CloseSource()
{
  if (twain_source_opened) // Must be in state 4
  {
    TW_UINT16 rc;

    rc = (*pDSM_Entry)(&AppId, NULL, DG_CONTROL, DAT_IDENTITY, MSG_CLOSEDS, (TW_MEMREF)&SelId);

    if (rc == TWRC_SUCCESS)
    {
      // Now back in state 3.
      twain_source_opened = false;
      ZeroMemory(&SelId, sizeof(SelId));
    }
    // Otherwise error, as above.
  }
}
//---------------------------------------------------------------------------
// DoAcquire
// Kicks off the process of image acquisition, using the above routines.
static void DoAcquire(TForm *form)
{
  if (twain_opened && !twain_source_opened) // Must be in state 3; Must not be in state 4 already.
  {
    OpenSource();

    if (twain_source_opened) // Now in state 4.
    {
      NegotiateCapabilities();

      if (twain_caps_negotiated) // Need capabilities
      {
        EnableSource(form);

        if (!twain_source_enabled) // If not in state 5, drop back to state 3
          CloseSource();
      }
    }
  }
}
//---------------------------------------------------------------------------
// AppMessage
// TWAIN requires all events to be passed down to it while a source is active, so it can see if it wants to handle any
// of them itself. If rc is TWRC_NOTDSEVENT then it's a regular old Windows event and should be processed by the default
// message handling in TApplication::ProcessMessages.
void __fastcall TForm1::AppMessage(tagMSG &Msg, bool &Handled)
{
  TW_EVENT twEvent;
  TW_INT16 rc;

  rc = TWRC_NOTDSEVENT;

  if (twain_source_enabled) // In state 5 or higher - a source is enabled and needs events
  {
    twEvent.pEvent = (TW_MEMREF)&Msg;
    twEvent.TWMessage = MSG_NULL;

    rc = (*pDSM_Entry)(&AppId, &SelId, DG_CONTROL, DAT_EVENT, MSG_PROCESSEVENT, (TW_MEMREF)&twEvent);

    switch (twEvent.TWMessage)
    {
    case MSG_XFERREADY:  // transfer is ready
      SetupAndTransferImage(this); // Now in state 6 courtesy of Source; Acknowledge and move to state 7 for acquisition.
      break;
    case MSG_CLOSEDSREQ: // request to close source
    case MSG_CLOSEDSOK:  // source closed (? - verify)
      DisableSource();
      if (!twain_source_enabled)
      {
        CloseSource();
        if (!twain_source_opened)
          menuFileAcquire->Enabled = true;
      }
      break;
    case MSG_NULL: // nothing.
    default:
      break;
    }
  }
  if (rc == TWRC_NOTDSEVENT) // not handled?
    ; // In Prometheus, other custom event handling shall be done here (see TformPrometheusMenu::HandleApplicationMessages)
  else
    Handled = true; // Tell VCL that TWAIN handled this event
}
//---------------------------------------------------------------------------
// jhaley: TWAIN prototype interface code ends here ^^^^^
//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
        : TForm(Owner)
{
  // Makes drawing much faster
  ScrollBox1->DoubleBuffered = true;

  // Load the source manager
  LoadSourceManager();

  if (twain_loaded)
  {
    // Source manager loaded; setup the message loop handler
    Application->OnMessage = AppMessage;

    // Open the source manager
    OpenSourceManager(this);
  }
}
//---------------------------------------------------------------------------
void __fastcall TForm1::menuFileExitClick(TObject *Sender)
{
  Close();
}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormClose(TObject *Sender, TCloseAction &Action)
{
  if (twain_loaded) // In state 2?
  {
    if (twain_opened) // In state 3?
    {
      if (twain_source_opened) // In state 4?
      {
        if (twain_source_enabled) // In state 5 or higher?
        {
          // Try to disable TWAIN Data Source (State 5 -> 4)
          DisableSource();

          if (twain_source_enabled) // STILL in state 5? Abort all pending transfers and try again.
          {
            DoAbortXfer();
            DisableSource();
          }
        }
        
        // Close TWAIN Data Source (State 4 -> 3)
        CloseSource();
      }

      // Close TWAIN DSM (State 3 -> 2)
      HWND handle = Handle;
      (*pDSM_Entry)(&AppId, NULL, DG_CONTROL, DAT_PARENT, MSG_CLOSEDSM, (TW_MEMREF)&handle);
      twain_opened = false;
    }

    // Shut down TWAIN DSM (State 2 -> 1)
    if (hDSMLib != NULL)
      FreeLibrary(hDSMLib);
    twain_loaded = false;
  }
}
//---------------------------------------------------------------------------
void __fastcall TForm1::menuFileSelectSourceClick(TObject *Sender)
{
  SelectSource();
}
//---------------------------------------------------------------------------
void __fastcall TForm1::menuFileAcquireClick(TObject *Sender)
{
  menuFileAcquire->Enabled = false;
  DoAcquire(this);
}
//---------------------------------------------------------------------------
void __fastcall TForm1::menuViewNextImageClick(TObject *Sender)
{
  DisplayNextBitmap(*imgBitmap);
}
//---------------------------------------------------------------------------
void __fastcall TForm1::menuViewPrevImageClick(TObject *Sender)
{
  DisplayPrevBitmap(*imgBitmap);
}
//---------------------------------------------------------------------------

